import cv2
import math
import numpy as np
import logging

logger = logging.getLogger(__name__)


class ROISelectionWin(object):
    def __init__(self):
        self._selected_point = None  # point used to edit the polygon vertice
        self._start_point = None  # square starting point
        self._end_point = None  # square ending point

    def circlebtn_click_evt(self):
        if self._circlebtn.checked:
            self._roi += ["[]"]
            self._roi.selected_row_index = len(self._roi) - 1
            self._circlebtn.checked = True
            self._polybtn.checked = False
            self._rectbtn.checked = False
            self._add_points_btn.checked = False

    def polybtn_click_evt(self):
        if self._polybtn.checked:
            self._roi += ["[]"]
            self._roi.selected_row_index = len(self._roi) - 1
            self._polybtn.checked = True
            self._circlebtn.checked = False
            self._rectbtn.checked = False
            self._add_points_btn.checked = False

    def rectbtn_click_evt(self):
        if self._rectbtn.checked:
            self._polybtn.checked = False
            self._circlebtn.checked = False
            self._rectbtn.checked = True
            self._add_points_btn.checked = False

    def create_mask(self, height, width):
        """
        Create a mask based on the selected ROIs
        """
        mask = np.zeros((height, width), dtype=np.uint8)

        if len(self._roi) > 0:
            for row in self._roi.value:
                if isinstance(row, str):
                    points = eval(row)
                elif isinstance(row[0], str):
                    points = eval(row[0])
                else:
                    points = row
                if len(points) < 3:
                    continue
                mask = cv2.fillPoly(
                    mask, [np.array(points, np.int32)], (1, 1, 1)
                )
        else:
            mask = mask + 1
        return mask

    def roi_selection_changed_evt(self):
        """
        Function called when the selected polygon change
        """
        self._polybtn.checked = False
        self._circlebtn.checked = False
        self._rectbtn.checked = False
        self._add_points_btn.checked = False

    def remove_roi(self):
        self._roi -= -1
        self._roi.selected_row_index = None  # No poly is selected
        self._player.refresh()

    def draw_rois(self, frame):
        """
        Draw the ROIs lines in the frame.
        """
        # Draw polygons
        try:
            index = self._roi.selected_row_index

            for row_index, row in enumerate(self._roi.value):
                points = eval(row[0])

                if len(points) > 3:
                    cv2.polylines(
                        frame,
                        [np.array(points, np.int32)],
                        True,
                        (0, 255, 0),
                        2,
                        lineType=cv2.LINE_AA,
                    )
                else:
                    # draw vertices
                    for i, point in enumerate(points):
                        point = tuple(point)
                        if self._selected_point == i and index == row_index:
                            cv2.circle(frame, point, 4, (0, 0, 255), 2)
                        else:
                            cv2.circle(frame, point, 4, (0, 255, 0), 2)

                if (
                    index == row_index
                    and self._selected_point is not None
                    and len(points) > self._selected_point
                ):
                    point = tuple(points[self._selected_point])
                    cv2.circle(frame, point, 4, (0, 0, 255), 2)

        except Exception as e:
            logger.debug(str(e), exc_info=True)

        # Draw the polygons in edition
        if self._start_point and self._end_point:

            if self._rectbtn.checked:
                cv2.rectangle(
                    frame, self._start_point, self._end_point, (233, 44, 44), 1
                )

        return frame



import sys, os

sys.path.append(os.getcwd())


def start():
    import logging, locale, coloredlogs

    logger = logging.getLogger(__name__)
    logging.basicConfig(
        filename="idtrackerai-app.log",
        filemode="w",
        format="%(name)s - %(levelname)s - %(message)s",
    )
    coloredlogs.install(
        level="DEBUG",
        fmt="[%(levelname)-8s] %(name)-40s %(message)s",
        # stream=open("idtrackerai-app.log", 'w')
    )
    coloredlogs.install(
        level="DEBUG",
        fmt="[%(levelname)-8s] %(name)-40s %(message)s",
        stream=open("idtrackerai-app.log", "w"),
    )

    # logging.getLogger('').addHandler(open("idtrackerai-app.log", 'w'))

    from pyforms import start_app
    from confapp import conf

    logging.getLogger("PyQt5").setLevel(logging.INFO)
    logging.getLogger("matplotlib").setLevel(logging.INFO)

    if conf.PYFORMS_MODE == "GUI":
        from .win_idtrackerai import IdTrackerAiGUI as App
    else:
        from .base_idtrackerai import BaseIdTrackerAi as App

    try:
        start_app(App, geometry=(100, 100, 800, 600))
    except SystemExit:
        pass
    except Exception as e:
        logger.info(e, exc_info=True)
        import sys
        import traceback

        ex_type, ex, tb = sys.exc_info()
        traceback.print_exception(ex_type, ex, tb)


# Execute the application
if __name__ == "__main__":
    start()



import os, subprocess, cv2, numpy as np, sys
import glob
from confapp import conf

from PyQt5.QtWidgets import QApplication
from AnyQt.QtWidgets import QMessageBox
from scipy import ndimage  # TODO: used to fill binary holes see below

from idtrackerai.video import Video
from idtrackerai.animals_detection.segmentation import _process_frame
from idtrackerai.postprocessing.individual_videos import (
    generate_individual_videos,
)
from idtrackerai.postprocessing.trajectories_to_video import (
    generate_trajectories_video,
)

from pyforms.controls import ControlButton
from pyforms.controls import ControlPlayer
from pyforms.controls import ControlCheckBox

from .base_idtrackerai import BaseIdTrackerAi
from .gui.grapharea_win import GraphAreaWin
from .gui.range_win import RangeWin

NEW_GUI_FORMSET = (
    [
        ("_video", "_session", "_savebtn"),
        ("_number_of_animals", "_toggle_blobs_area_info", "_chcksegm", " "),
        ("_intensity", "_bgsub"),
        ("_area", "_resreduct"),
        ("_range", "_rangelst", "_addrange", "_multiple_range"),
        ("_applyroi", " "),
        ("_rectbtn", "_polybtn", "_circlebtn", " "),
        "_roi",
        ("_no_ids", "_pre_processing", "_progress", "_validation"),
        ("_indiv_videos", "_traj_video", " "),
    ],
    "_player",
)
OLD_GUI_FORMSET = [
    ("_video", "_session", "_savebtn"),
    "_player",
    "=",
    ("_number_of_animals", "_toggle_blobs_area_info", "_chcksegm", " "),
    ("_intensity", "_bgsub"),
    ("_area", "_resreduct"),
    ("_range", "_rangelst", "_addrange", "_multiple_range"),
    ("_applyroi", " "),
    ("_rectbtn", "_polybtn", "_circlebtn", " "),
    "_roi",
    ("_no_ids", "_pre_processing", "_progress", "_validation"),
    ("_indiv_videos", "_traj_video", " "),
]


class IdTrackerAiGUI(BaseIdTrackerAi):

    INDIV_VIDEO_BTN_LABEL = "Generate individual videos"
    GEN_INDIV_VIDEO_BTN_LABEL = "Generating individual videos..."
    TRAJ_VIDEO_BTN_LABEL = "Generate video with trajectories"
    GEN_TRAJ_VIDEO_BTN_LABEL = "Generate video with trajectories..."
    FINAL_MESSAGE_DEFAULT = "idtracker.ai terminated."

    def __init__(self, *args, **kwargs):

        self._player = ControlPlayer(
            "Player", enabled=False, multiple_files=True
        )
        self._toggle_blobs_area_info = ControlCheckBox(
            "Segmented blobs info",
            changed_event=self.__toggle_graph_evt,
            enabled=False,
        )
        self._pre_processing = ControlButton(
            "Track video", default=self.track_video, enabled=False
        )
        self._savebtn = ControlButton(
            "Save parameters", default=self.save_window, enabled=False
        )

        self._polybtn = ControlButton(
            "Polygon",
            checkable=True,
            enabled=False,
            default=self.polybtn_click_evt,
        )
        self._rectbtn = ControlButton(
            "Rectangle",
            checkable=True,
            enabled=False,
            default=self.rectbtn_click_evt,
        )
        self._circlebtn = ControlButton(
            "Ellipse",
            checkable=True,
            enabled=False,
            default=self.circlebtn_click_evt,
        )

        self._add_points_btn = ControlButton(
            "Add setup points",
            checkable=True,
            enabled=False,
            default=self.add_points_dict_click_evt,
        )

        self._addrange = ControlButton(
            "Add interval", default=self.__rangelst_add_evt, visible=False
        )

        self._indiv_videos = ControlButton(
            self.INDIV_VIDEO_BTN_LABEL,
            default=self.__generate_individual_videos_evt,
            enabled=False,
        )
        self._traj_video = ControlButton(
            self.TRAJ_VIDEO_BTN_LABEL,
            default=self.__generate_trajectories_video_evt,
            enabled=False,
        )
        self._validation = ControlButton(
            "Validate trajectories",
            default=self.__open_videoannotator_evt,
            enabled=False,
        )

        self._final_message = self.FINAL_MESSAGE_DEFAULT

        self._graph = GraphAreaWin(parent_win=self)

        super().__init__(*args, **kwargs)

        self.set_margin(10)

        self.formset = (
            [
                ("_video", "_resreduct", " "),
                (
                    "_number_of_animals",
                    "_toggle_blobs_area_info",
                    "_chcksegm",
                    " ",
                ),
                "_intensity",
                "_area",
                ("_applyroi", "_bgsub", " "),
                ("_rectbtn", "_polybtn", "_circlebtn", " "),
                "_roi",
                ("_range", "_rangelst", "_addrange", "_multiple_range"),
                ("_add_setup_info", " "),
                ("_add_points_btn", " "),
                "_points_list",
                ("_session", "_savebtn", " "),
                ("_no_ids", "_pre_processing", "_progress"),
                ("_validation", "_traj_video", "_indiv_videos", " "),
            ],
            "_player",
        )

        self._graph.on_draw = self.__graph_on_draw_evt
        self._applyroi.changed_event = self.__apply_roi_changed_evt
        self._add_setup_info.changed_event = self.__add_setup_info_changed_evt
        self._session.changed_event = self.__session_changed_evt
        self._player.drag_event = self.on_player_drag_in_video_window
        self._player.end_drag_event = self.on_player_end_drag_in_video_window
        self._player.click_event = self.on_player_click_in_video_window
        self._video.changed_event = self.__video_changed_evt
        self._video_path.changed_event = self.__video_changed_evt
        self._intensity.changed_event = self._player.refresh
        self._player.double_click_event = (
            self.on_player_double_click_in_video_window
        )
        self._player.process_frame_event = self.process_frame_evt
        self._multiple_range.changed_event = self.__multiple_range_changed_evt
        self._resreduct.changed_event = self.__resreduct_changed_evt

        self.setMinimumHeight(conf.GUI_MINIMUM_HEIGHT)
        self.setMinimumWidth(conf.GUI_MINIMUM_WIDTH)
        self._video.form.setMinimumWidth(300)
        self._player.setMinimumHeight(300)
        self._player.setMinimumWidth(300)
        self._toggle_blobs_area_info.form.setMaximumWidth(180)
        self._roi.setMaximumHeight(100)
        self._session.form.setMaximumWidth(250)
        self._multiple_range.form.setMaximumWidth(130)

        self.__apply_roi_changed_evt()
        self.__add_setup_info_changed_evt()
        self.__session_changed_evt()

    def save_form(self, data={}, path=None):
        data["open-multiple-files"] = self.open_multiple_files
        return super().save_form(data, path)

    def track_video(self):
        super().track_video()
        self.info(self._final_message, "Ended")

    def set_controls_enabled(self, status):
        super().set_controls_enabled(status)
        self._validation.enabled = status
        self._indiv_videos.enabled = status
        self._traj_video.enabled = status
        self._pre_processing.enabled = status
        self._savebtn.enabled = status
        self._player.enabled = status
        self._polybtn.enabled = status
        self._rectbtn.enabled = status
        self._circlebtn.enabled = status
        self._toggle_blobs_area_info.enabled = status
        self._addrange.enabled = status

        self._add_points_btn.enabled = status

    def process_frame_evt(self, frame):
        """
        Function called before an image is shown in the player.
        It does the pre-visualization segmentation and ROIs selection.
        """
        # Save original shape to rescale if resolution reduction is applied
        original_size = frame.shape[1], frame.shape[0]  # (width, height)
        self._frame_width = original_size[0]
        self._frame_height = original_size[1]
        # TODO: check if bkgmodel needs to be updated because of new ROI
        self._mask_img = self.create_mask(
            self._frame_height, self._frame_width
        )
        animal_detection_parameters = {
            "number_of_animals": int(self._number_of_animals.value),
            "min_threshold": self._intensity.value[0],
            "max_threshold": self._intensity.value[1],
            "min_area": self._area.value[0],
            "max_area": self._area.value[1],
            "tracking_interval": None,
            "apply_ROI": self._applyroi.value,
            "rois": self._roi.value,
            "mask": self._mask_img,
            "subtract_bkg": self._bgsub.value,
            "bkg_model": self._background_img,
            "resolution_reduction": self._resreduct.value,
            "sigma_gaussian_blurring": conf.SIGMA_GAUSSIAN_BLURRING,
        }

        (boxes, _, _, areas, _, contours, _,) = _process_frame(
            frame,
            animal_detection_parameters,
            -1,  # Get frame_number from the Widget
            save_pixels="NONE",
            save_segmentation_image="NONE",
        )

        # Save detected areas to plot the bar plot of the blobs size in pixels
        self._detected_areas = areas
        # Update graph with areas of seglemted blobs
        if conf.PYFORMS_MODE == "GUI" and self._toggle_blobs_area_info.value:
            self._graph.draw()
        # Draw detected blobs in frame
        if animal_detection_parameters["resolution_reduction"] != 1:
            frame = cv2.resize(
                frame,
                None,
                fx=animal_detection_parameters["resolution_reduction"],
                fy=animal_detection_parameters["resolution_reduction"],
                interpolation=cv2.INTER_AREA,
            )
        cv2.drawContours(frame, contours, -1, color=(0, 0, 255), thickness=-1)
        # Resize to original size (ROI and setup points are in original size)
        frame = cv2.resize(frame, original_size, interpolation=cv2.INTER_AREA)
        # Draw ROIs in frame
        self.draw_rois(frame)
        # Draw setup points in frame
        self.draw_points_list(frame)
        return frame

    def __resreduct_changed_evt(self):
        self._player.refresh()

    def __rangelst_add_evt(self):
        win = RangeWin(
            parent_win=self,
            maximum=self._player.max,
            begin=self._player.video_index,
            control_list=self._rangelst,
        )
        win.show()

    def __multiple_range_changed_evt(self):
        if self._multiple_range.value:
            self._range.hide()
            self._addrange.show()
            self._rangelst.show()
        else:
            self._range.show()
            self._addrange.hide()
            self._rangelst.hide()

    def __video_changed_evt(self):
        """
        Function called when the video file is selected.
        Ask to the player to load the video file.
        """
        if self.video_path:
            self._player.value = self.video_path
            if self._player.value:
                self._range.max = self._player.max
                self._range.value = [0, self._player.max]
                self.set_controls_enabled(True)
                self._player.forward_one_frame()
            else:
                self.set_controls_enabled(False)
            video = Video(self.video_path, self.open_multiple_files)
            self.video_object = video
        else:
            self.set_controls_enabled(False)

    def bgsub_changed_evt(self):
        super().bgsub_changed_evt()
        self._player.refresh()

    def __toggle_graph_evt(self):
        if self._toggle_blobs_area_info.value:
            self._graph.show()
        else:
            self._graph.hide()

    def __graph_on_draw_evt(self, figure):
        areas = self._detected_areas

        if not areas:
            return
        figure.gca().remove()
        axes = figure.add_subplot(111)
        axes.clear()
        axes.bar(range(1, len(areas) + 1), areas, 0.5)
        if len(areas) > 0:
            min_area = np.min(areas)
            axes.axhline(min_area, color="k", linewidth=0.3)
            axes.set_title(
                str(len(areas))
                + " blobs detected. Minimum area: "
                + str(min_area)
            )
            axes.set_xlabel("Blob index")
            axes.set_ylabel("Area in pixels")

    def __apply_roi_changed_evt(self):
        """
        Hide and show the controls to setup the background
        """
        if self._applyroi.value:
            self._roi.show()
            self._polybtn.show()
            self._rectbtn.show()
            self._circlebtn.show()
        else:
            self._roi.hide()
            self._polybtn.hide()
            self._rectbtn.hide()
            self._circlebtn.hide()

    def __add_setup_info_changed_evt(self):
        """
        Hide and show the controls to setup the background
        """
        if self._add_setup_info.value:
            self._points_list.show()
            self._add_points_btn.show()
        else:
            self._points_list.hide()
            self._add_points_btn.hide()

    def __session_changed_evt(self):
        video_folder = os.path.dirname(self.video_path)
        session_folder = "session_{0}".format(self._session.value)
        session_path = os.path.join(video_folder, session_folder)
        blobs_path_1 = os.path.join(
            session_path, "preprocessing", "blobs_collection_no_gaps.npy"
        )
        blobs_path_2 = os.path.join(
            session_path, "preprocessing", "blobs_collection.npy"
        )
        vidobj_path = os.path.join(session_path, "video_object.npy")
        trajs_wo_path = os.path.join(session_path, "trajectories_wo_gaps")
        trajs_path = os.path.join(session_path, "trajectories")

        if (
            os.path.exists(session_path)
            and (os.path.exists(blobs_path_1) or os.path.exists(blobs_path_2))
            and os.path.exists(vidobj_path)
        ):
            self._validation.enabled = True
        else:
            self._validation.enabled = False

        if (
            os.path.exists(session_path)
            and (os.path.exists(trajs_path) or os.path.exists(trajs_wo_path))
            and os.path.exists(vidobj_path)
        ):
            self._indiv_videos.enabled = True
            self._traj_video.enabled = True
        else:
            self._indiv_videos.enabled = False
            self._traj_video.enabled = False

    def __update_progress_evt(self, progress_count, max_count=None):
        if max_count is not None:
            self._progress.max = max_count
            self._progress.value = 0
            self._progress.show()
        elif self._progress.max == progress_count:
            self._progress.value = 0
        else:
            self._progress.value = progress_count

    def __open_videoannotator_evt(self):

        video_folder = os.path.dirname(self.video_path)
        session_folder = "session_{0}".format(self._session.value)
        session_path = os.path.join(video_folder, session_folder)

        subprocess.Popen(
            [sys.executable, "-m", "pythonvideoannotator", session_path]
        )

    def __generate_individual_videos_evt(self):
        self._indiv_videos.enabled = False
        self._indiv_videos.label = self.GEN_INDIV_VIDEO_BTN_LABEL
        QApplication.processEvents()
        video_object, trajectories = get_video_object_and_trajectories(
            self.video_path, self._session.value
        )
        generate_individual_videos(video_object, trajectories)
        self._indiv_videos.enabled = True
        self._indiv_videos.label = self.INDIV_VIDEO_BTN_LABEL
        QApplication.processEvents()

    def __generate_trajectories_video_evt(self):
        self._traj_video.enabled = False
        self._traj_video.label = self.GEN_TRAJ_VIDEO_BTN_LABEL
        QApplication.processEvents()
        video_object, trajectories = get_video_object_and_trajectories(
            self.video_path, self._session.value
        )
        generate_trajectories_video(video_object, trajectories)
        self._traj_video.enabled = True
        self._traj_video.label = self.TRAJ_VIDEO_BTN_LABEL
        QApplication.processEvents()

    @property
    def open_multiple_files(self):
        return self._player.multiple_files


def get_video_object_and_trajectories(video_path, session_name):
    video_folder = os.path.dirname(video_path)
    session_folder = "session_{0}".format(session_name)
    session_path = os.path.join(video_folder, session_folder)
    trajs_wo_path = os.path.join(session_path, "trajectories_wo_gaps")
    trajs_path = os.path.join(session_path, "trajectories")
    traj_wo_ids_path = os.path.join(
        session_path, "trajectories_wo_identification"
    )

    video_object = np.load(
        os.path.join(session_path, "video_object.npy"), allow_pickle=True
    ).item()

    if os.path.exists(trajs_wo_path):
        trajectories_file = glob.glob(os.path.join(trajs_wo_path, "*.npy"))[-1]
    elif os.path.exists(trajs_path):
        trajectories_file = glob.glob(os.path.join(trajs_path, "*.npy"))[-1]
    elif os.path.exists(traj_wo_ids_path):
        trajectories_file = glob.glob(os.path.join(traj_wo_ids_path, "*.npy"))[
            -1
        ]

    trajectories = np.load(trajectories_file, allow_pickle=True).item()[
        "trajectories"
    ]

    return video_object, trajectories



import math
import numpy as np
import logging
import cv2

logger = logging.getLogger(__name__)


def points_distance(p1, p2):
    return math.hypot(p2[0] - p1[0], p2[1] - p1[1])


class SetupInfoWin(object):
    def __init__(self):
        self._selected_point = None  # point used to edit the polygon vertice

    def add_points_dict_click_evt(self):
        # Initialize list
        if self._add_points_btn.checked:
            name = self.input_text(
                "Input name of list of points to be added",
                title="List of points name",
                default=None,
            )
            self._points_list += ['["{}", ]'.format(name)]
            self._points_list.selected_row_index = len(self._points_list) - 1
            self._add_points_btn.checked = True
            self._polybtn.checked = False
            self._circlebtn.checked = False
            self._rectbtn.checked = False

    def add_setup_info_changed_evt(self):
        """
        Function called when the selected polygon change
        """
        self._add_points_btn.checked = False
        self._polybtn.checked = False
        self._circlebtn.checked = False
        self._rectbtn.checked = False

    def remove_setup_info(self):
        self._points_list -= -1
        self._points_list.selected_row_index = None  # No poly is selected
        self._player.refresh()

    def draw_points_list(self, frame):
        """
        Draw the ROIs lines in the frame.
        """
        # Draw selected points
        try:
            index = self._points_list.selected_row_index

            for row_index, row in enumerate(self._points_list.value):
                points = eval(row[0])
                # draw vertices
                for i, point in enumerate(points):
                    if not isinstance(point, str):
                        point = tuple(point)
                        if self._selected_point == i and index == row_index:
                            cv2.circle(frame, point, 4, (0, 0, 255), 2)
                        else:
                            cv2.circle(frame, point, 4, (0, 255, 0), 2)

        except Exception as e:
            logger.debug(str(e), exc_info=True)

        return frame

    def remove_points_list(self):
        self._points_list -= -1
        self._points_list.selected_row_index = None  # No poly is selected
        self._player.refresh()

    def create_setup_poitns_dict(self):
        if len(self._points_list) > 0:
            points_dict = {}
            for row in self._points_list.value:

                if isinstance(row, str):
                    points = eval(row)
                elif isinstance(row[0], str):
                    points = eval(row[0])
                else:
                    points = row

                points_dict[points[0]] = np.asarray(points[1:])

            return points_dict
        else:
            return None



from pyforms.basewidget import BaseWidget
from pyforms.controls import ControlMatplotlib


class GraphAreaWin(BaseWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(title="Area graph")

        self._graph = ControlMatplotlib(
            "Blobs area", toolbar=False, enabled=False
        )

        self.formset = ["_graph"]

    @property
    def enabled(self):
        return self._graph.enabled

    @enabled.setter
    def enabled(self, value):
        self._graph.enabled = value

    @property
    def on_draw(self):
        return self._graph.on_draw

    @on_draw.setter
    def on_draw(self, value):
        self._graph.on_draw = value

    def draw(self):
        self._graph.draw()



import numpy as np, os, logging

from confapp import conf

from pyforms.basewidget import BaseWidget
from pyforms.controls import ControlText
from pyforms.controls import ControlCheckBox
from pyforms.controls import ControlFile
from pyforms.controls import ControlList
from pyforms.controls import ControlBoundingSlider

from pyforms.controls import ControlNumber
from pyforms.controls import ControlProgress

from idtrackerai.animals_detection.segmentation_utils import (
    compute_background,
)

from idtrackerai.video import Video


from idtrackerai.animals_detection import AnimalsDetectionAPI
from idtrackerai.crossings_detection import CrossingsDetectionAPI
from idtrackerai.fragmentation import FragmentationAPI
from idtrackerai.tracker.tracker import TrackerAPI

from .gui.roi_selection import ROISelectionWin
from .gui.setup_points_selection import SetupInfoWin
from .gui.player_win_interactions import PlayerWinInteractions


logger = logging.getLogger(__name__)
try:
    import local_settings

    conf += local_settings
except ImportError:
    logger.info("Local settings file not available.")


class BaseIdTrackerAi(
    BaseWidget, PlayerWinInteractions, ROISelectionWin, SetupInfoWin
):

    ERROR_MESSAGE_DEFAULT = (
        "\n \nIf this error persists please open an issue at "
        "https://gitlab.com/polavieja_lab/idtrackerai or "
        "send an email to idtrackerai@gmail.com. "
        "Check the log file idtrackerai-app.log in your "
        "working directory and attach it to the issue."
    )

    SEGMENTATION_CHECK_FINAL_MESSAGE = (
        "Readjust the segmentation parameters and track the video again."
    )

    def __init__(self, *args, **kwargs):
        BaseWidget.__init__(self, title="idtracker.ai")
        ROISelectionWin.__init__(self)
        SetupInfoWin.__init__(self)

        # Main objects of idtracker.ai
        self.video_object = None
        self.list_of_blobs = None
        self.list_of_fragments = None
        self.list_of_global_fragments = None

        # App user interaction items
        # Session folder
        self._session = ControlText("Session", default="test")
        # Path to the video
        self._video = ControlFile("Video")
        self._video_path = ControlFile(
            "Video file. Note: overwrite the _video "
            "parameter defined in the json. "
            "Nice to have to execute the "
            "application in a cluster environment"
        )
        # Apply a ROI option
        self._applyroi = ControlCheckBox("Apply ROI", enabled=False)
        # List of ROIs that together will build the mask
        self._roi = ControlList(
            "ROI",
            enabled=False,
            readonly=True,
            select_entire_row=True,
            item_selection_changed_event=self.roi_selection_changed_evt,
            remove_function=self.remove_roi,
        )
        # Apply background subtraction option
        self._bgsub = ControlCheckBox(
            "Subtract background",
            changed_event=self.bgsub_changed_evt,
            enabled=False,
        )
        # Check that the number of segmented blobs in each frame is equal or
        # lower than the number of animals in the video
        self._chcksegm = ControlCheckBox("Check segmentation", enabled=False)
        # Resolution reduction factor
        self._resreduct = ControlNumber(
            "Resolution reduction",
            default=conf.RES_REDUCTION_DEFAULT,
            minimum=0.01,
            maximum=1,
            decimals=2,
            step=0.1,
            enabled=False,
        )
        # Intensity thresholds for animal detection (segmentation)
        self._intensity = ControlBoundingSlider(
            "Intensity thresholds",
            default=[conf.MIN_THRESHOLD_DEFAULT, conf.MAX_THRESHOLD_DEFAULT],
            min=conf.MIN_THRESHOLD,
            max=conf.MAX_THRESHOLD,
            enabled=False,
        )
        # Area thresholds for animal detection (segmentation)
        self._area = ControlBoundingSlider(
            "Area thresholds",
            default=[conf.MIN_AREA_DEFAULT, conf.MAX_AREA_DEFAULT],
            min=conf.AREA_LOWER,
            max=conf.AREA_UPPER,
            enabled=False,
        )
        # Ranges of frames to track
        self._range = ControlBoundingSlider(
            "Tracking interval",
            default=[0, 508],
            min=0,
            max=255,
            enabled=False,
        )  ###TODO: Upate max to the number of frames in the video
        self._rangelst = ControlText("Tracking intervals", visible=False)
        self._multiple_range = ControlCheckBox("Multiple", enabled=False)
        # Number of animals in the video
        self._number_of_animals = ControlNumber(
            "Number of animals",
            default=conf.NUMBER_OF_ANIMALS_DEFAULT,
            enabled=False,
        )
        # Tracking without identification
        self._no_ids = ControlCheckBox("Track without identities")
        # Setup points to store together with the trajectories
        self._add_setup_info = ControlCheckBox("Add setup info", enabled=False)
        self._points_list = ControlList(
            "Setup info",
            enabled=False,
            readonly=True,
            select_entire_row=True,
            item_selection_changed_event=self.add_setup_info_changed_evt,
            remove_function=self.remove_setup_info,
        )
        # Progress bar
        self._progress = ControlProgress("Progress", enabled=False)

        # App attributes
        self.formset = [
            ("_video", "_session"),
            ("_range", "_rangelst", "_multiple_range"),
            "_intensity",
            "_area",
            (
                "_number_of_animals",
                "_resreduct",
                " ",
                "_applyroi",
                "_chcksegm",
                "_bgsub",
            ),
            "_roi",
            ("_no_ids", "_progress"),
            ("_add_setup_info", " "),
            "_points_list",
        ]

        self.load_order = [
            "_session",
            "_video",
            "_range",
            "_rangelst",
            "_multiple_range",
            "_intensity",
            "_area",
            "_number_of_animals",
            "_resreduct",
            "_chcksegm",
            "_roi",
            "_no_ids",
            "_applyroi",
            "_bgsub",
            "_add_setup_info",
            "_points_list",
        ]

        # store the computed background with the correct resolution reduction
        self._background_img = None
        self._mask_img = None

        # flag to open multiple video files with similar names
        self._multiple_files = False

        self._frame_height = None
        self._frame_width = None

    #########################################################
    ## GUI EVENTS ###########################################
    #########################################################

    def load_form(self, data, path=None):
        self._multiple_files = data.get("open-multiple-files", False)
        super().load_form(data, path)

    def bgsub_changed_evt(self):
        if self.video_object is not None:
            self.__get_bkg_model()

    def set_controls_enabled(self, status):
        self._applyroi.enabled = status
        self._roi.enabled = status
        self._bgsub.enabled = status
        self._chcksegm.enabled = status
        self._resreduct.enabled = status
        self._intensity.enabled = status
        self._area.enabled = status
        self._range.enabled = status
        self._number_of_animals.enabled = status
        self._progress.enabled = status
        self._multiple_range.enabled = status
        self._rangelst.enabled = status
        self._no_ids.enabled = status
        self._add_setup_info.enabled = status
        self._points_list.enabled = status

    def track_video(self):
        logger.info("Calling track_video")
        self._video.enabled = False
        self._session.enabled = False
        self.set_controls_enabled(False)

        try:
            # Init tracking manager
            self._step0_init_video_object()
            self._step1_get_user_defined_parameters()
            # Preprocessing
            # success will be False if there are more blobs than animals and
            # the user asked to check the segmentation consistency
            success = self._step2_pre_processing()
            # Training and identification and post processing
            if success:
                success = self._step3_tracking()
            if success:
                # This flag is important to register the smoke tests that work
                logger.info("Success")

        except Exception as e:
            self.save()
            logger.error(e, exc_info=True)
            self.critical(str(e), "Error")

        self._video.enabled = True
        self._session.enabled = True
        self.set_controls_enabled(True)

    def save(self):
        logger.info("Saving objects from base_idtrackerai")
        self.video_object.save()
        if self.list_of_blobs is not None:
            self.list_of_blobs.save(self.video_object.blobs_path)
        if self.list_of_fragments is not None:
            self.list_of_fragments.save(self.video_object.fragments_path)
        if self.list_of_global_fragments is not None:
            self.list_of_global_fragments.save(
                self.video_object.global_fragments_path,
                self.list_of_fragments.fragments,
            )

    def _step0_init_video_object(self):

        self._progress.max = 4

        if not os.path.exists(self.video_path):
            raise Exception(
                "The video you are trying to track does not exist or the "
                f"path to the video is wrong: {self.video_path}"
            )

        # INIT AND POPULATE VIDEO OBJECT WITH PARAMETERS
        if self.video_object is None:
            logger.info("START: INIT VIDEO OBJECT")
            self.video_object = Video(
                video_path=self.video_path,
                open_multiple_files=self.open_multiple_files,
            )
            logger.info("FINISH: INIT VIDEO OBJECT")
        self.video_object.create_session_folder(self._session.value)

    def __get_tracking_interval(self):
        if self._multiple_range.value and self._rangelst.value:
            try:
                self._tracking_interval = eval(self._rangelst.value)
            except Exception as e:
                logger.fatal(e, exc_info=True)
                self._tracking_interval = [self._range.value]
        else:
            self._tracking_interval = [self._range.value]

    def __get_bkg_model(self):
        if self._bgsub.value:
            if self._background_img is None:
                # Asked for background subtraction but it is not computed
                logger.info("Computing the background model")
                self._mask_img = self.create_mask(
                    self.video_object.original_height,
                    self.video_object.original_width,
                )
                self._background_img = compute_background(
                    self.video_object.video_paths,
                    self.video_object.original_height,
                    self.video_object.original_width,
                    self.video_object.video_path,
                    self._mask_img,
                    self.video_object.episodes_start_end,
                )
            else:
                logger.info("Storing previously computed background model")
        else:
            # Did not ask for background subtraction
            logger.info("No background model computed")
            self._background_img = None

    def __get_mask(self):
        if self._applyroi:
            self._mask_img = self.create_mask(
                self.video_object.original_height,
                self.video_object.original_width,
            )
        else:
            self._mask_img = np.ones(
                (
                    self.video_object.original_height,
                    self.video_object.original_width,
                )
            )

    def _step1_get_user_defined_parameters(self):
        # TODO: Make background subtraction depend on the tracking interval
        # Collect tracking interval before computing the background
        self.__get_tracking_interval()
        self.__get_mask()
        # The computation of the background has a computation of the mask
        # TODO: Separate better mask and bkg
        self.__get_bkg_model()
        # TODO: Separate user defined parameters and advanced parameters
        # There are other parameters that come form the local_settings.py
        # It would be great to store them all in a singe json file so we
        # can check all the parameters used for tracking
        user_defined_parameters = {
            "number_of_animals": int(self._number_of_animals.value),
            "min_threshold": self._intensity.value[0],
            "max_threshold": self._intensity.value[1],
            "min_area": self._area.value[0],
            "max_area": self._area.value[1],
            "check_segmentation": self._chcksegm.value,
            "tracking_interval": self._tracking_interval,
            "apply_ROI": self._applyroi.value,
            "rois": self._roi.value,
            "mask": self._mask_img,
            "subtract_bkg": self._bgsub.value,
            "bkg_model": self._background_img,
            "resolution_reduction": self._resreduct.value,
            "track_wo_identification": self._no_ids.value,
            "setup_points": self.create_setup_poitns_dict(),
            "sigma_gaussian_blurring": conf.SIGMA_GAUSSIAN_BLURRING,
            "knowledge_transfer_folder": conf.KNOWLEDGE_TRANSFER_FOLDER_IDCNN,
            "identity_transfer": False,
            "identification_image_size": None,
        }

        if conf.IDENTITY_TRANSFER:
            # TODO: the identification_image_size is not really passed by
            # the used but inferred from the knowledge transfer folder
            (
                user_defined_parameters["identity_transfer"],
                user_defined_parameters["identification_image_size"],
            ) = TrackerAPI.check_if_identity_transfer_is_possible(
                user_defined_parameters["number_of_animals"],
                conf.KNOWLEDGE_TRANSFER_FOLDER_IDCNN,
            )

        self.video_object._user_defined_parameters = user_defined_parameters

    def __output_segmentation_consistency_warning(self, outfile_path):
        self.warning(
            "On some frames it was found more blobs than "
            "animals, "
            "you can find the index of these frames in the file:"
            f"<p>{outfile_path}</p>"
            "<p>Please readjust the segmentation parameters and "
            "press 'Track video' again.</p>",
            "Found more blobs than animals",
        )
        self._progress.value = 0
        self._final_message = self.SEGMENTATION_CHECK_FINAL_MESSAGE

    def _step2_pre_processing(self):

        logger.info("START: ANIMAL DETECTION")
        animals_detector = AnimalsDetectionAPI(self.video_object)
        self.list_of_blobs = animals_detector()
        # Check segmentation consistency
        segmentation_consistent = animals_detector.check_segmentation()
        if not segmentation_consistent and self._chcksegm.value:
            outfile_path = animals_detector.save_inconsistent_frames()
            self.save()  # saves video_object
            self.__output_segmentation_consistency_warning(outfile_path)
            return False  # This will make the tracking finish
        self._progress.value = 1
        logger.info("FINISH: ANIMAL DETECTION")

        logger.info("START: CROSSING DETECTION")
        crossings_detector = CrossingsDetectionAPI(
            self.video_object, self.list_of_blobs
        )
        crossings_detector()
        self._progress.value = 2
        logger.info("FINISH: CROSSING DETECTION")

        logger.info("START: FRAGMENTATION")
        fragmentator = FragmentationAPI(self.video_object, self.list_of_blobs)
        self.list_of_fragments = fragmentator()
        self._progress.value = 3
        logger.info("FINISH: FRAGMENTATION")
        return True  # This will make the tracking continue

    def _step3_tracking(self):

        tracker = TrackerAPI(
            self.video_object, self.list_of_blobs, self.list_of_fragments
        )

        if self.video_object.user_defined_parameters[
            "track_wo_identification"
        ]:
            # START: FRAGMENTATION
            logger.info("START: TRACKING WITHOUT IDENTITIES")
            tracker.track_wo_identification()
            logger.info("FINISH: TRACKING WITHOUT IDENTITIES")
            self._final_message = (
                "Tracking without identities finished. "
                "No estimated accuracy computed."
            )
        else:
            if (
                self.video_object.user_defined_parameters["number_of_animals"]
                == 1
            ):
                logger.info("START: TRACKING SINGLE ANIMAL")
                tracker.track_single_animal()
                logger.info("FINISH: TRACKING SINGLE ANIMAL")

            else:
                tracker.track_multiple_animals()
                self.list_of_fragments.update_identification_images_dataset()

            logger.info(
                "Estimated accuracy: {}".format(
                    self.video_object.estimated_accuracy
                )
            )

            self.video_object.delete_data()

            self._final_message = (
                "Tracking finished with {0:.2f} "
                "estimated accuracy.".format(
                    self.video_object.estimated_accuracy * 100
                )
            )
        self._progress.value = 4
        return True

    def __update_progress(self, value, label=None, total=None):

        if total is not None:
            self._progress.max = total
        if label is not None:
            self._progress.label = label

        self._progress.value = value

    def alert(self, msg, title=None):
        msg = msg + self.ERROR_MESSAGE_DEFAULT
        self.message(msg, title, msg_type="error")

    def critical(self, msg, title=None):
        msg = msg + self.ERROR_MESSAGE_DEFAULT
        self.message(msg, title, msg_type="error")

    @property
    def open_multiple_files(self):
        return self._multiple_files

    @property
    def video_path(self):
        return (
            self._video_path.value
            if self._video_path.value
            else self._video.value
        )



import cv2
import math
import numpy as np
import logging

logger = logging.getLogger(__name__)


def points_distance(p1, p2):
    return math.hypot(p2[0] - p1[0], p2[1] - p1[1])


def create_rectangle_points(start, end):
    return [start, (end[0], start[1]), end, (start[0], end[1])]


def get_intersection_point_distance(test_point, point1, point2):
    p1 = np.float32(point1)
    p2 = np.float32(point2)
    p3 = np.float32(test_point)
    dist = np.linalg.norm(np.cross(p2 - p1, p1 - p3)) / np.linalg.norm(p2 - p1)
    return dist


class PlayerWinInteractions(object):
    def __init__(self):
        self._selected_point = None  # point used to edit the polygon vertice
        self._start_point = None  # square starting point
        self._end_point = None  # square ending point

    def select_point(self, x, y):
        """
        select a point from the selected polygon

        :param x: X coordenate
        :param y: Y coordenate
        :return: Index of the selected point in the polygon
        """
        if (
            self._polybtn.checked
            or self._circlebtn.checked
            or self._rectbtn.checked
        ):
            points_list = self._roi
        else:
            points_list = self._points_list
        index = points_list.selected_row_index

        if index is not None:
            try:
                coord = x, y
                points = eval(points_list.value[index][0])
                for i, point in enumerate(points):
                    if points_distance(coord, point) <= 5:
                        self._selected_point = i
                        return
                self._selected_point = None
            except Exception as e:
                print(e)
                pass

    def on_player_double_click_in_video_window(self, event, x, y):
        """
        Event called when the mouse double click.
        This event is used to create vertices on polygons edges.
        :param event: Mouse event
        :param x: X coordenate
        :param y: Y coordenate
        """
        mouse = int(x), int(y)
        distances = []

        index = self._roi.selected_row_index

        # if a polygon is selected
        if index is not None:
            try:
                points = list(eval(self._roi.value[index][0]))
                n_points = len(points)
                for point_index, point in enumerate(points):
                    next_point = points[(point_index + 1) % n_points]
                    distance = get_intersection_point_distance(
                        mouse, point, next_point
                    )

                    if distance <= 5:
                        vector = (
                            next_point[0] - point[0],
                            next_point[1] - point[1],
                        )
                        center = (
                            point[0] + vector[0] / 2,
                            point[1] + vector[1] / 2,
                        )
                        radius = points_distance(center, point)
                        mouse_distance = points_distance(center, mouse)
                        if mouse_distance < radius:
                            distances.append((distance, point_index))
            ## TODO: Do not use bare except
            except Exception as e:
                print(e)
                pass

            if len(distances) > 0:
                distances = sorted(distances, key=lambda x: x[0])
                point_index = distances[0][1]
                points.insert(point_index + 1, mouse)

                self._roi.set_value(0, index, str(points)[1:-1])

                self._selected_point = point_index + 1

                if not self._player.is_playing:
                    self._player.refresh()

    def on_player_click_in_video_window(self, event, x, y):
        """
        Called when the mouse click in the player
        If the poly button is active it should add the clicked point to the active polygon
        :param event: Mouse event
        :param x: X coordinate
        :param y: Y coordinate
        """
        mouse = int(x), int(y)
        self._selected_point = None
        if (
            not self._rectbtn.checked
            and not self._polybtn.checked
            and not self._circlebtn.checked
            and not self._add_points_btn.checked
        ):
            self.select_point(*mouse)

        if (
            self._polybtn.checked
            or self._circlebtn.checked
            or self._add_points_btn.checked
        ):
            if self._polybtn.checked or self._circlebtn.checked:
                points_list = self._roi
            else:
                points_list = self._points_list
            index = points_list.selected_row_index

            if index is not None:
                points = list(eval(points_list.value[index][0]))
                points.append(mouse)
                if len(points) >= 5 and self._circlebtn.checked:
                    center, axis, angle = cv2.fitEllipse(
                        np.array(points, np.int32)
                    )
                    center = int(round(center[0])), int(round(center[1]))
                    axis = int(round(axis[0] / 2.0)), int(round(axis[1] / 2.0))
                    angle = int(round(angle))
                    points = cv2.ellipse2Poly(center, axis, angle, 0, 360, 5)
                    points_list.set_value(
                        0, index, str(points.tolist())[1:-1] + ","
                    )
                    self._circlebtn.checked = False
                else:
                    points_list.set_value(0, index, str(points)[1:-1] + ",")

    def on_player_drag_in_video_window(self, start_point, end_point):
        """
        Called when the mouse drag start
        :param start_point: Top left point
        :param end_point: Bottom right point
        """

        self._start_point = int(start_point[0]), int(start_point[1])
        self._end_point = int(end_point[0]), int(end_point[1])

        if self._selected_point is not None:
            index = self._roi.selected_row_index
            if index is not None:
                try:
                    points = list(eval(self._roi.value[index][0]))
                    points[self._selected_point] = self._end_point
                    self._roi.set_value(0, index, str(points)[1:-1] + ",")
                except Exception as e:
                    print(e)

        # Refresh the image if the video is not playing
        if not self._player.is_playing:
            self._player.refresh()

    def on_player_end_drag_in_video_window(self, start_point, end_point):
        """
        Called when the mouse drag ends
        :param start_point: Top left point
        :param end_point: Bottom right point
        """
        # Draw the rectangle
        if self._rectbtn.checked:
            self._start_point = int(start_point[0]), int(start_point[1])
            self._end_point = int(end_point[0]), int(end_point[1])
            points = create_rectangle_points(
                self._start_point, self._end_point
            )
            self._roi += [str(points)[1:-1]]
            self._start_point = None
            self._end_point = None
            self._rectbtn.checked = False

        # Refresh the image if the video is not playing
        if not self._player.is_playing:
            self._player.refresh()



from pyforms.basewidget import BaseWidget
from pyforms.controls import ControlNumber
from pyforms.controls import ControlButton


class RangeWin(BaseWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(title="Frames range")

        self._list = kwargs.get("control_list", None)

        self._begin = ControlNumber(
            "Begin",
            default=kwargs.get("begin", 0),
            maximum=kwargs.get("maximum", 100),
        )
        self._end = ControlNumber(
            "End",
            default=kwargs.get("begin", 0),
            maximum=kwargs.get("maximum", 100),
        )
        self._okbtn = ControlButton("Ok", default=self.__okbtn_evt)

        self.formset = [("_begin", "_end"), "_okbtn"]

        self.set_margin(10)

    def __okbtn_evt(self):
        if self._begin.value > self._end.value:
            self.alert(
                "The <b>begin</b> frame has to be smaller than <b>end</b> frame."
            )
        else:
            value = str([int(self._begin.value), int(self._end.value)])
            if self._list.value:
                self._list.value += ", " + value
            else:
                self._list.value = value
            self.close()



#!/usr/bin/python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
import re

version = ""
with open("idtrackerai_app/__init__.py", "r") as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    ).group(1)

setup(
    name="idtrackerai-app",
    version=version,
    description="""""",
    author=["Ricardo Ribeiro", "Francisco Romero Ferrero"],
    author_email="idtrackerai@gmail.com, ricardojvr@gmail.com, paco.romero.ferrero@gmail.com",
    url="https://idtrackerai-app.readthedocs.org",
    packages=find_packages(),
    install_requires=[
        "coloredlogs",
        # "idtrackerai == 4.0.0a0",
        "pyforms-terminal==4.*",
    ],
    extras_require={
        "gui": [
            "pyforms-gui==4.904.152",
            "python-video-annotator==3.306",
            "python-video-annotator-module-idtrackerai",  # ==1.0.0a0
        ],
        "only-gui": ["pyforms-gui==4.904.152"],
    },
    entry_points={
        "console_scripts": [
            "idtrackerai=idtrackerai_app.__main__:start",
        ],
    },
)
